@@ -13,19 +13,14 @@
 #include <time.h>
 #include <assert.h>
 #include <unistd.h>
+#include <poll.h>
 
 /* Forward Declarations */
 static void item_link_q(item *it);
 static void item_unlink_q(item *it);
 
-#define HOT_LRU 0
-#define WARM_LRU 64
-#define COLD_LRU 128
-#define NOEXP_LRU 192
 static unsigned int lru_type_map[4] = {HOT_LRU, WARM_LRU, COLD_LRU, NOEXP_LRU};
 
-#define CLEAR_LRU(id) (id & ~(3<<6))
-
 #define LARGEST_ID POWER_LARGEST
 typedef struct {
     uint64_t evicted;
@@ -33,8 +28,8 @@ typedef struct {
     uint64_t reclaimed;
     uint64_t outofmemory;
     uint64_t tailrepairs;
-    uint64_t expired_unfetched;
-    uint64_t evicted_unfetched;
+    uint64_t expired_unfetched; /* items reclaimed but never touched */
+    uint64_t evicted_unfetched; /* items evicted but never touched */
     uint64_t crawler_reclaimed;
     uint64_t crawler_items_checked;
     uint64_t lrutail_reflocked;
@@ -45,35 +40,21 @@ typedef struct {
     rel_time_t evicted_time;
 } itemstats_t;
 
-typedef struct {
-    uint64_t histo[61];
-    uint64_t ttl_hourplus;
-    uint64_t noexp;
-    uint64_t reclaimed;
-    uint64_t seen;
-    rel_time_t start_time;
-    rel_time_t end_time;
-    bool run_complete;
-} crawlerstats_t;
-
 static item *heads[LARGEST_ID];
 static item *tails[LARGEST_ID];
-static crawler crawlers[LARGEST_ID];
 static itemstats_t itemstats[LARGEST_ID];
 static unsigned int sizes[LARGEST_ID];
-static crawlerstats_t crawlerstats[MAX_NUMBER_OF_SLAB_CLASSES];
+static uint64_t sizes_bytes[LARGEST_ID];
+static unsigned int *stats_sizes_hist = NULL;
+static uint64_t stats_sizes_cas_min = 0;
+static int stats_sizes_buckets = 0;
 
-static int crawler_count = 0;
-static volatile int do_run_lru_crawler_thread = 0;
 static volatile int do_run_lru_maintainer_thread = 0;
-static int lru_crawler_initialized = 0;
 static int lru_maintainer_initialized = 0;
 static int lru_maintainer_check_clsid = 0;
-static pthread_mutex_t lru_crawler_lock = PTHREAD_MUTEX_INITIALIZER;
-static pthread_cond_t  lru_crawler_cond = PTHREAD_COND_INITIALIZER;
-static pthread_mutex_t lru_crawler_stats_lock = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t lru_maintainer_lock = PTHREAD_MUTEX_INITIALIZER;
 static pthread_mutex_t cas_id_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t stats_sizes_lock = PTHREAD_MUTEX_INITIALIZER;
 
 void item_stats_reset(void) {
     int i;
@@ -84,9 +65,19 @@ void item_stats_reset(void) {
     }
 }
 
+/* called with class lru lock held */
+void do_item_stats_add_crawl(const int i, const uint64_t reclaimed,
+        const uint64_t unfetched, const uint64_t checked) {
+    itemstats[i].crawler_reclaimed += reclaimed;
+    itemstats[i].expired_unfetched += unfetched;
+    itemstats[i].crawler_items_checked += checked;
+}
+
+#define LRU_PULL_EVICT 1
+#define LRU_PULL_CRAWL_BLOCKS 2
+
 static int lru_pull_tail(const int orig_id, const int cur_lru,
-        const unsigned int total_chunks, const bool do_evict, const uint32_t cur_hv);
-static int lru_crawler_start(uint32_t id, uint32_t remaining);
+        const uint64_t total_bytes, uint8_t flags);
 
 /* Get the next CAS id for a new item. */
 /* TODO: refactor some atomics for this. */
@@ -116,7 +107,7 @@ static unsigned int noexp_lru_size(int slabs_clsid) {
     id |= NOEXP_LRU;
     unsigned int ret;
     pthread_mutex_lock(&lru_locks[id]);
-    ret = sizes[id];
+    ret = sizes_bytes[id];
     pthread_mutex_unlock(&lru_locks[id]);
     return ret;
 }
@@ -144,21 +135,19 @@ static unsigned int noexp_lru_size(int slabs_clsid) {
  *
  * Returns the total size of the header.
  */
-static size_t item_make_header(const uint8_t nkey, const int flags, const int nbytes,
+static size_t item_make_header(const uint8_t nkey, const unsigned int flags, const int nbytes,
                      char *suffix, uint8_t *nsuffix) {
     /* suffix is defined at 40 chars elsewhere.. */
-    *nsuffix = (uint8_t) snprintf(suffix, 40, " %d %d\r\n", flags, nbytes - 2);
+    *nsuffix = (uint8_t) snprintf(suffix, 40, " %u %d\r\n", flags, nbytes - 2);
     return sizeof(item) + nkey + *nsuffix + nbytes;
 }
 
-item *do_item_alloc(char *key, const size_t nkey, const int flags,
-                    const rel_time_t exptime, const int nbytes,
-                    const uint32_t cur_hv) {
+item *do_item_alloc(char *key, const size_t nkey, const unsigned int flags,
+                    const rel_time_t exptime, const int nbytes) {
     int i;
     uint8_t nsuffix;
     item *it = NULL;
     char suffix[40];
-    unsigned int total_chunks;
     size_t ntotal = item_make_header(nkey + 1, flags, nbytes, suffix, &nsuffix);
     if (settings.use_cas) {
         ntotal += sizeof(uint64_t);
@@ -174,21 +163,34 @@ item *do_item_alloc(char *key, const size_t nkey, const int flags,
      * occasional OOM's, rather than internally work around them.
      * This also gives one fewer code path for slab alloc/free
      */
-    for (i = 0; i < 5; i++) {
+    /* TODO: if power_largest, try a lot more times? or a number of times
+     * based on how many chunks the new object should take up?
+     * or based on the size of an object lru_pull_tail() says it evicted?
+     * This is a classical GC problem if "large items" are of too varying of
+     * sizes. This is actually okay here since the larger the data, the more
+     * bandwidth it takes, the more time we can loop in comparison to serving
+     * and replacing small items.
+     */
+    for (i = 0; i < 10; i++) {
+        uint64_t total_bytes;
         /* Try to reclaim memory first */
         if (!settings.lru_maintainer_thread) {
-            lru_pull_tail(id, COLD_LRU, 0, false, cur_hv);
+            lru_pull_tail(id, COLD_LRU, 0, 0);
         }
-        it = slabs_alloc(ntotal, id, &total_chunks, 0);
+        it = slabs_alloc(ntotal, id, &total_bytes, 0);
+
         if (settings.expirezero_does_not_evict)
-            total_chunks -= noexp_lru_size(id);
+            total_bytes -= noexp_lru_size(id);
+
         if (it == NULL) {
             if (settings.lru_maintainer_thread) {
-                lru_pull_tail(id, HOT_LRU, total_chunks, false, cur_hv);
-                lru_pull_tail(id, WARM_LRU, total_chunks, false, cur_hv);
-                lru_pull_tail(id, COLD_LRU, total_chunks, true, cur_hv);
+                lru_pull_tail(id, HOT_LRU, total_bytes, 0);
+                lru_pull_tail(id, WARM_LRU, total_bytes, 0);
+                if (lru_pull_tail(id, COLD_LRU, total_bytes, LRU_PULL_EVICT) <= 0)
+                    break;
             } else {
-                lru_pull_tail(id, COLD_LRU, 0, true, cur_hv);
+                if (lru_pull_tail(id, COLD_LRU, 0, LRU_PULL_EVICT) <= 0)
+                    break;
             }
         } else {
             break;
@@ -212,7 +214,8 @@ item *do_item_alloc(char *key, const size_t nkey, const int flags,
     //assert(it != heads[id]);
 
     /* Refcount is seeded to 1 by slabs_alloc() */
-    it->next = it->prev = it->h_next = 0;
+    it->next = it->prev = 0;
+
     /* Items are initially loaded into the HOT_LRU. This is '0' but I want at
      * least a note here. Compiler (hopefully?) optimizes this out.
      */
@@ -229,13 +232,29 @@ item *do_item_alloc(char *key, const size_t nkey, const int flags,
     it->slabs_clsid = id;
 
     DEBUG_REFCNT(it, '*');
-    it->it_flags = settings.use_cas ? ITEM_CAS : 0;
+    it->it_flags |= settings.use_cas ? ITEM_CAS : 0;
     it->nkey = nkey;
     it->nbytes = nbytes;
     memcpy(ITEM_key(it), key, nkey);
     it->exptime = exptime;
     memcpy(ITEM_suffix(it), suffix, (size_t)nsuffix);
     it->nsuffix = nsuffix;
+
+    /* Need to shuffle the pointer stored in h_next into it->data. */
+    if (it->it_flags & ITEM_CHUNKED) {
+        item_chunk *chunk = (item_chunk *) ITEM_data(it);
+
+        chunk->next = (item_chunk *) it->h_next;
+        chunk->prev = 0;
+        chunk->head = it;
+        /* Need to chain back into the head's chunk */
+        chunk->next->prev = chunk;
+        chunk->size = chunk->next->size - ((char *)chunk - (char *)it);
+        chunk->used = 0;
+        assert(chunk->size > 0);
+    }
+    it->h_next = 0;
+
     return it;
 }
 
@@ -284,6 +303,7 @@ static void do_item_link_q(item *it) { /* item is the new head */
     *head = it;
     if (*tail == 0) *tail = it;
     sizes[it->slabs_clsid]++;
+    sizes_bytes[it->slabs_clsid] += it->nbytes;
     return;
 }
 
@@ -312,6 +332,7 @@ static void do_item_unlink_q(item *it) {
     if (it->next) it->next->prev = it->prev;
     if (it->prev) it->prev->next = it->next;
     sizes[it->slabs_clsid]--;
+    sizes_bytes[it->slabs_clsid] -= it->nbytes;
     return;
 }
 
@@ -328,8 +349,8 @@ int do_item_link(item *it, const uint32_t hv) {
     it->time = current_time;
 
     STATS_LOCK();
-    stats.curr_bytes += ITEM_ntotal(it);
-    stats.curr_items += 1;
+    stats_state.curr_bytes += ITEM_ntotal(it);
+    stats_state.curr_items += 1;
     stats.total_items += 1;
     STATS_UNLOCK();
 
@@ -338,6 +359,7 @@ int do_item_link(item *it, const uint32_t hv) {
     assoc_insert(it, hv);
     item_link_q(it);
     refcount_incr(&it->refcount);
+    item_stats_sizes_add(it);
 
     return 1;
 }
@@ -347,9 +369,10 @@ void do_item_unlink(item *it, const uint32_t hv) {
     if ((it->it_flags & ITEM_LINKED) != 0) {
         it->it_flags &= ~ITEM_LINKED;
         STATS_LOCK();
-        stats.curr_bytes -= ITEM_ntotal(it);
-        stats.curr_items -= 1;
+        stats_state.curr_bytes -= ITEM_ntotal(it);
+        stats_state.curr_items -= 1;
         STATS_UNLOCK();
+        item_stats_sizes_remove(it);
         assoc_delete(ITEM_key(it), it->nkey, hv);
         item_unlink_q(it);
         do_item_remove(it);
@@ -362,9 +385,10 @@ void do_item_unlink_nolock(item *it, const uint32_t hv) {
     if ((it->it_flags & ITEM_LINKED) != 0) {
         it->it_flags &= ~ITEM_LINKED;
         STATS_LOCK();
-        stats.curr_bytes -= ITEM_ntotal(it);
-        stats.curr_items -= 1;
+        stats_state.curr_bytes -= ITEM_ntotal(it);
+        stats_state.curr_items -= 1;
         STATS_UNLOCK();
+        item_stats_sizes_remove(it);
         assoc_delete(ITEM_key(it), it->nkey, hv);
         do_item_unlink_q(it);
         do_item_remove(it);
@@ -617,6 +641,73 @@ void item_stats(ADD_STAT add_stats, void *c) {
     add_stats(NULL, 0, NULL, 0, c);
 }
 
+bool item_stats_sizes_status(void) {
+    bool ret = false;
+    mutex_lock(&stats_sizes_lock);
+    if (stats_sizes_hist != NULL)
+        ret = true;
+    mutex_unlock(&stats_sizes_lock);
+    return ret;
+}
+
+void item_stats_sizes_init(void) {
+    if (stats_sizes_hist != NULL)
+        return;
+    stats_sizes_buckets = settings.item_size_max / 32 + 1;
+    stats_sizes_hist = calloc(stats_sizes_buckets, sizeof(int));
+    stats_sizes_cas_min = (settings.use_cas) ? get_cas_id() : 0;
+}
+
+void item_stats_sizes_enable(ADD_STAT add_stats, void *c) {
+    mutex_lock(&stats_sizes_lock);
+    if (!settings.use_cas) {
+        APPEND_STAT("sizes_status", "error", "");
+        APPEND_STAT("sizes_error", "cas_support_disabled", "");
+    } else if (stats_sizes_hist == NULL) {
+        item_stats_sizes_init();
+        if (stats_sizes_hist != NULL) {
+            APPEND_STAT("sizes_status", "enabled", "");
+        } else {
+            APPEND_STAT("sizes_status", "error", "");
+            APPEND_STAT("sizes_error", "no_memory", "");
+        }
+    } else {
+        APPEND_STAT("sizes_status", "enabled", "");
+    }
+    mutex_unlock(&stats_sizes_lock);
+}
+
+void item_stats_sizes_disable(ADD_STAT add_stats, void *c) {
+    mutex_lock(&stats_sizes_lock);
+    if (stats_sizes_hist != NULL) {
+        free(stats_sizes_hist);
+        stats_sizes_hist = NULL;
+    }
+    APPEND_STAT("sizes_status", "disabled", "");
+    mutex_unlock(&stats_sizes_lock);
+}
+
+void item_stats_sizes_add(item *it) {
+    if (stats_sizes_hist == NULL || stats_sizes_cas_min > ITEM_get_cas(it))
+        return;
+    int ntotal = ITEM_ntotal(it);
+    int bucket = ntotal / 32;
+    if ((ntotal % 32) != 0) bucket++;
+    if (bucket < stats_sizes_buckets) stats_sizes_hist[bucket]++;
+}
+
+/* I think there's no way for this to be accurate without using the CAS value.
+ * Since items getting their time value bumped will pass this validation.
+ */
+void item_stats_sizes_remove(item *it) {
+    if (stats_sizes_hist == NULL || stats_sizes_cas_min > ITEM_get_cas(it))
+        return;
+    int ntotal = ITEM_ntotal(it);
+    int bucket = ntotal / 32;
+    if ((ntotal % 32) != 0) bucket++;
+    if (bucket < stats_sizes_buckets) stats_sizes_hist[bucket]--;
+}
+
 /** dumps out a list of objects of each size, with granularity of 32 bytes */
 /*@null@*/
 /* Locks are correct based on a technicality. Holds LRU lock while doing the
@@ -624,39 +715,23 @@ void item_stats(ADD_STAT add_stats, void *c) {
  * which don't change.
  */
 void item_stats_sizes(ADD_STAT add_stats, void *c) {
+    mutex_lock(&stats_sizes_lock);
 
-    /* max 1MB object, divided into 32 bytes size buckets */
-    const int num_buckets = 32768;
-    unsigned int *histogram = calloc(num_buckets, sizeof(int));
-
-    if (histogram != NULL) {
+    if (stats_sizes_hist != NULL) {
         int i;
-
-        /* build the histogram */
-        for (i = 0; i < LARGEST_ID; i++) {
-            pthread_mutex_lock(&lru_locks[i]);
-            item *iter = heads[i];
-            while (iter) {
-                int ntotal = ITEM_ntotal(iter);
-                int bucket = ntotal / 32;
-                if ((ntotal % 32) != 0) bucket++;
-                if (bucket < num_buckets) histogram[bucket]++;
-                iter = iter->next;
-            }
-            pthread_mutex_unlock(&lru_locks[i]);
-        }
-
-        /* write the buffer */
-        for (i = 0; i < num_buckets; i++) {
-            if (histogram[i] != 0) {
+        for (i = 0; i < stats_sizes_buckets; i++) {
+            if (stats_sizes_hist[i] != 0) {
                 char key[8];
                 snprintf(key, sizeof(key), "%d", i * 32);
-                APPEND_STAT(key, "%u", histogram[i]);
+                APPEND_STAT(key, "%u", stats_sizes_hist[i]);
             }
         }
-        free(histogram);
+    } else {
+        APPEND_STAT("sizes_status", "disabled", "");
     }
+
     add_stats(NULL, 0, NULL, 0, c);
+    mutex_unlock(&stats_sizes_lock);
 }
 
 /** wrapper around assoc_find which does the lazy expiration logic */
@@ -705,6 +780,9 @@ item *do_item_get(const char *key, const size_t nkey, const uint32_t hv, conn *c
             do_item_unlink(it, hv);
             do_item_remove(it);
             it = NULL;
+            pthread_mutex_lock(&c->thread->stats.mutex);
+            c->thread->stats.get_flushed++;
+            pthread_mutex_unlock(&c->thread->stats.mutex);
             if (settings.verbose > 2) {
                 fprintf(stderr, " -nuked by flush");
             }
@@ -748,7 +826,7 @@ item *do_item_touch(const char *key, size_t nkey, uint32_t exptime,
 /* Returns number of items remove, expired, or evicted.
  * Callable from worker threads or the LRU maintainer thread */
 static int lru_pull_tail(const int orig_id, const int cur_lru,
-        const unsigned int total_chunks, const bool do_evict, const uint32_t cur_hv) {
+        const uint64_t total_bytes, uint8_t flags) {
     item *it = NULL;
     int id = orig_id;
     int removed = 0;
@@ -760,7 +838,7 @@ static int lru_pull_tail(const int orig_id, const int cur_lru,
     item *next_it;
     void *hold_lock = NULL;
     unsigned int move_to_lru = 0;
-    uint64_t limit;
+    uint64_t limit = 0;
 
     id |= cur_lru;
     pthread_mutex_lock(&lru_locks[id]);
@@ -771,13 +849,17 @@ static int lru_pull_tail(const int orig_id, const int cur_lru,
         next_it = search->prev;
         if (search->nbytes == 0 && search->nkey == 0 && search->it_flags == 1) {
             /* We are a crawler, ignore it. */
+            if (flags & LRU_PULL_CRAWL_BLOCKS) {
+                pthread_mutex_unlock(&lru_locks[id]);
+                return 0;
+            }
             tries++;
             continue;
         }
         uint32_t hv = hash(ITEM_key(search), search->nkey);
         /* Attempt to hash item lock the "search" item. If locked, no
          * other callers can incr the refcount. Also skip ourselves. */
-        if (hv == cur_hv || (hold_lock = item_trylock(hv)) == NULL)
+        if ((hold_lock = item_trylock(hv)) == NULL)
             continue;
         /* Now see if the item is refcount locked */
         if (refcount_incr(&search->refcount) != 2) {
@@ -820,10 +902,11 @@ static int lru_pull_tail(const int orig_id, const int cur_lru,
          */
         switch (cur_lru) {
             case HOT_LRU:
-                limit = total_chunks * settings.hot_lru_pct / 100;
+                limit = total_bytes * settings.hot_lru_pct / 100;
             case WARM_LRU:
-                limit = total_chunks * settings.warm_lru_pct / 100;
-                if (sizes[id] > limit) {
+                if (limit == 0)
+                    limit = total_bytes * settings.warm_lru_pct / 100;
+                if (sizes_bytes[id] > limit) {
                     itemstats[id].moves_to_cold++;
                     move_to_lru = COLD_LRU;
                     do_item_unlink_q(search);
@@ -844,7 +927,7 @@ static int lru_pull_tail(const int orig_id, const int cur_lru,
                 break;
             case COLD_LRU:
                 it = search; /* No matter what, we're stopping */
-                if (do_evict) {
+                if (flags & LRU_PULL_EVICT) {
                     if (settings.evict_to_free == 0) {
                         /* Don't think we need a counter for this. It'll OOM.  */
                         break;
@@ -903,14 +986,14 @@ static int lru_maintainer_juggle(const int slabs_clsid) {
     int i;
     int did_moves = 0;
     bool mem_limit_reached = false;
-    unsigned int total_chunks = 0;
+    uint64_t total_bytes = 0;
     unsigned int chunks_perslab = 0;
     unsigned int chunks_free = 0;
     /* TODO: if free_chunks below high watermark, increase aggressiveness */
     chunks_free = slabs_available_chunks(slabs_clsid, &mem_limit_reached,
-            &total_chunks, &chunks_perslab);
+            &total_bytes, &chunks_perslab);
     if (settings.expirezero_does_not_evict)
-        total_chunks -= noexp_lru_size(slabs_clsid);
+        total_bytes -= noexp_lru_size(slabs_clsid);
 
     /* If slab automove is enabled on any level, and we have more than 2 pages
      * worth of chunks free in this class, ask (gently) to reassign a page
@@ -923,11 +1006,11 @@ static int lru_maintainer_juggle(const int slabs_clsid) {
     /* Juggle HOT/WARM up to N times */
     for (i = 0; i < 1000; i++) {
         int do_more = 0;
-        if (lru_pull_tail(slabs_clsid, HOT_LRU, total_chunks, false, 0) ||
-            lru_pull_tail(slabs_clsid, WARM_LRU, total_chunks, false, 0)) {
+        if (lru_pull_tail(slabs_clsid, HOT_LRU, total_bytes, LRU_PULL_CRAWL_BLOCKS) ||
+            lru_pull_tail(slabs_clsid, WARM_LRU, total_bytes, LRU_PULL_CRAWL_BLOCKS)) {
             do_more++;
         }
-        do_more += lru_pull_tail(slabs_clsid, COLD_LRU, total_chunks, false, 0);
+        do_more += lru_pull_tail(slabs_clsid, COLD_LRU, total_bytes, LRU_PULL_CRAWL_BLOCKS);
         if (do_more == 0)
             break;
         did_moves++;
@@ -949,20 +1032,22 @@ static int lru_maintainer_juggle(const int slabs_clsid) {
  * The latter is to avoid newly started daemons from waiting too long before
  * retrying a crawl.
  */
-static void lru_maintainer_crawler_check(void) {
+static void lru_maintainer_crawler_check(struct crawler_expired_data *cdata) {
     int i;
-    static rel_time_t last_crawls[MAX_NUMBER_OF_SLAB_CLASSES];
+    static rel_time_t next_crawls[MAX_NUMBER_OF_SLAB_CLASSES];
     static rel_time_t next_crawl_wait[MAX_NUMBER_OF_SLAB_CLASSES];
+    uint8_t todo[MAX_NUMBER_OF_SLAB_CLASSES];
+    memset(todo, 0, sizeof(uint8_t) * MAX_NUMBER_OF_SLAB_CLASSES);
+    bool do_run = false;
+    if (!cdata->crawl_complete) {
+        return;
+    }
+
     for (i = POWER_SMALLEST; i < MAX_NUMBER_OF_SLAB_CLASSES; i++) {
-        crawlerstats_t *s = &crawlerstats[i];
+        crawlerstats_t *s = &cdata->crawlerstats[i];
         /* We've not successfully kicked off a crawl yet. */
-        if (last_crawls[i] == 0) {
-            if (lru_crawler_start(i, 0) > 0) {
-                last_crawls[i] = current_time;
-            }
-        }
-        pthread_mutex_lock(&lru_crawler_stats_lock);
         if (s->run_complete) {
+            pthread_mutex_lock(&cdata->lock);
             int x;
             /* Should we crawl again? */
             uint64_t possible_reclaims = s->seen - s->noexp;
@@ -974,27 +1059,45 @@ static void lru_maintainer_crawler_check(void) {
             rel_time_t since_run = current_time - s->end_time;
             /* Don't bother if the payoff is too low. */
             if (settings.verbose > 1)
-                fprintf(stderr, "maint crawler: low_watermark: %llu, possible_reclaims: %llu, since_run: %u\n",
-                        (unsigned long long)low_watermark, (unsigned long long)possible_reclaims,
+                fprintf(stderr, "maint crawler[%d]: low_watermark: %llu, possible_reclaims: %llu, since_run: %u\n",
+                        i, (unsigned long long)low_watermark, (unsigned long long)possible_reclaims,
                         (unsigned int)since_run);
             for (x = 0; x < 60; x++) {
-                if (since_run < (x * 60) + 60)
-                    break;
                 available_reclaims += s->histo[x];
+                if (available_reclaims > low_watermark) {
+                    if (next_crawl_wait[i] < (x * 60)) {
+                        next_crawl_wait[i] += 60;
+                    } else if (next_crawl_wait[i] >= 60) {
+                        next_crawl_wait[i] -= 60;
+                    }
+                    break;
+                }
             }
-            if (available_reclaims > low_watermark) {
-                last_crawls[i] = 0;
-                if (next_crawl_wait[i] > 60)
-                    next_crawl_wait[i] -= 60;
-            } else if (since_run > 5 && since_run > next_crawl_wait[i]) {
-                last_crawls[i] = 0;
-                if (next_crawl_wait[i] < MAX_MAINTCRAWL_WAIT)
-                    next_crawl_wait[i] += 60;
+
+            if (available_reclaims == 0) {
+                next_crawl_wait[i] += 60;
             }
+
+            if (next_crawl_wait[i] > MAX_MAINTCRAWL_WAIT) {
+                next_crawl_wait[i] = MAX_MAINTCRAWL_WAIT;
+            }
+
+            next_crawls[i] = current_time + next_crawl_wait[i] + 5;
             if (settings.verbose > 1)
-                fprintf(stderr, "maint crawler: available reclaims: %llu, next_crawl: %u\n", (unsigned long long)available_reclaims, next_crawl_wait[i]);
+                fprintf(stderr, "maint crawler[%d]: next_crawl: %u, [%d] now: [%d]\n",
+                        i, next_crawl_wait[i], next_crawls[i], current_time);
+            // Got our calculation, avoid running until next actual run.
+            s->run_complete = false;
+            pthread_mutex_unlock(&cdata->lock);
         }
-        pthread_mutex_unlock(&lru_crawler_stats_lock);
+        if (current_time > next_crawls[i]) {
+            todo[i] = 1;
+            do_run = true;
+            next_crawls[i] = current_time + 5; // minimum retry wait.
+        }
+    }
+    if (do_run) {
+        lru_crawler_start(todo, 0, CRAWLER_EXPIRED, cdata, NULL, 0);
     }
 }
 
@@ -1007,6 +1110,10 @@ static void *lru_maintainer_thread(void *arg) {
     int i;
     useconds_t to_sleep = MIN_LRU_MAINTAINER_SLEEP;
     rel_time_t last_crawler_check = 0;
+    struct crawler_expired_data cdata;
+    memset(&cdata, 0, sizeof(struct crawler_expired_data));
+    pthread_mutex_init(&cdata.lock, NULL);
+    cdata.crawl_complete = true; // kick off the crawler.
 
     pthread_mutex_lock(&lru_maintainer_lock);
     if (settings.verbose > 2)
@@ -1040,7 +1147,7 @@ static void *lru_maintainer_thread(void *arg) {
         }
         /* Once per second at most */
         if (settings.lru_crawler && last_crawler_check != current_time) {
-            lru_maintainer_crawler_check();
+            lru_maintainer_crawler_check(&cdata);
             last_crawler_check = current_time;
         }
     }
@@ -1099,16 +1206,15 @@ int init_lru_maintainer(void) {
     return 0;
 }
 
-/*** ITEM CRAWLER THREAD ***/
-
-static void crawler_link_q(item *it) { /* item is the new tail */
+/* Tail linkers and crawler for the LRU crawler. */
+void do_item_linktail_q(item *it) { /* item is the new tail */
     item **head, **tail;
     assert(it->it_flags == 1);
     assert(it->nbytes == 0);
 
     head = &heads[it->slabs_clsid];
     tail = &tails[it->slabs_clsid];
-    assert(*tail != 0);
+    //assert(*tail != 0);
     assert(it != *tail);
     assert((*head && *tail) || (*head == 0 && *tail == 0));
     it->prev = *tail;
@@ -1122,7 +1228,7 @@ static void crawler_link_q(item *it) { /* item is the new tail */
     return;
 }
 
-static void crawler_unlink_q(item *it) {
+void do_item_unlinktail_q(item *it) {
     item **head, **tail;
     head = &heads[it->slabs_clsid];
     tail = &tails[it->slabs_clsid];
@@ -1145,7 +1251,7 @@ static void crawler_unlink_q(item *it) {
 
 /* This is too convoluted, but it's a difficult shuffle. Try to rewrite it
  * more clearly. */
-static item *crawler_crawl_q(item *it) {
+item *do_item_crawl_q(item *it) {
     item **head, **tail;
     assert(it->it_flags == 1);
     assert(it->nbytes == 0);
@@ -1198,305 +1304,3 @@ static item *crawler_crawl_q(item *it) {
 
     return it->next; /* success */
 }
-
-/* I pulled this out to make the main thread clearer, but it reaches into the
- * main thread's values too much. Should rethink again.
- */
-static void item_crawler_evaluate(item *search, uint32_t hv, int i) {
-    int slab_id = CLEAR_LRU(i);
-    crawlerstats_t *s = &crawlerstats[slab_id];
-    itemstats[i].crawler_items_checked++;
-    if ((search->exptime != 0 && search->exptime < current_time)
-        || item_is_flushed(search)) {
-        itemstats[i].crawler_reclaimed++;
-        s->reclaimed++;
-
-        if (settings.verbose > 1) {
-            int ii;
-            char *key = ITEM_key(search);
-            fprintf(stderr, "LRU crawler found an expired item (flags: %d, slab: %d): ",
-                search->it_flags, search->slabs_clsid);
-            for (ii = 0; ii < search->nkey; ++ii) {
-                fprintf(stderr, "%c", key[ii]);
-            }
-            fprintf(stderr, "\n");
-        }
-        if ((search->it_flags & ITEM_FETCHED) == 0) {
-            itemstats[i].expired_unfetched++;
-        }
-        do_item_unlink_nolock(search, hv);
-        do_item_remove(search);
-        assert(search->slabs_clsid == 0);
-    } else {
-        s->seen++;
-        refcount_decr(&search->refcount);
-        if (search->exptime == 0) {
-            s->noexp++;
-        } else if (search->exptime - current_time > 3599) {
-            s->ttl_hourplus++;
-        } else {
-            rel_time_t ttl_remain = search->exptime - current_time;
-            int bucket = ttl_remain / 60;
-            s->histo[bucket]++;
-        }
-    }
-}
-
-static void *item_crawler_thread(void *arg) {
-    int i;
-    int crawls_persleep = settings.crawls_persleep;
-
-    pthread_mutex_lock(&lru_crawler_lock);
-    pthread_cond_signal(&lru_crawler_cond);
-    settings.lru_crawler = true;
-    if (settings.verbose > 2)
-        fprintf(stderr, "Starting LRU crawler background thread\n");
-    while (do_run_lru_crawler_thread) {
-    pthread_cond_wait(&lru_crawler_cond, &lru_crawler_lock);
-
-    while (crawler_count) {
-        item *search = NULL;
-        void *hold_lock = NULL;
-
-        for (i = POWER_SMALLEST; i < LARGEST_ID; i++) {
-            if (crawlers[i].it_flags != 1) {
-                continue;
-            }
-            pthread_mutex_lock(&lru_locks[i]);
-            search = crawler_crawl_q((item *)&crawlers[i]);
-            if (search == NULL ||
-                (crawlers[i].remaining && --crawlers[i].remaining < 1)) {
-                if (settings.verbose > 2)
-                    fprintf(stderr, "Nothing left to crawl for %d\n", i);
-                crawlers[i].it_flags = 0;
-                crawler_count--;
-                crawler_unlink_q((item *)&crawlers[i]);
-                pthread_mutex_unlock(&lru_locks[i]);
-                pthread_mutex_lock(&lru_crawler_stats_lock);
-                crawlerstats[CLEAR_LRU(i)].end_time = current_time;
-                crawlerstats[CLEAR_LRU(i)].run_complete = true;
-                pthread_mutex_unlock(&lru_crawler_stats_lock);
-                continue;
-            }
-            uint32_t hv = hash(ITEM_key(search), search->nkey);
-            /* Attempt to hash item lock the "search" item. If locked, no
-             * other callers can incr the refcount
-             */
-            if ((hold_lock = item_trylock(hv)) == NULL) {
-                pthread_mutex_unlock(&lru_locks[i]);
-                continue;
-            }
-            /* Now see if the item is refcount locked */
-            if (refcount_incr(&search->refcount) != 2) {
-                refcount_decr(&search->refcount);
-                if (hold_lock)
-                    item_trylock_unlock(hold_lock);
-                pthread_mutex_unlock(&lru_locks[i]);
-                continue;
-            }
-
-            /* Frees the item or decrements the refcount. */
-            /* Interface for this could improve: do the free/decr here
-             * instead? */
-            pthread_mutex_lock(&lru_crawler_stats_lock);
-            item_crawler_evaluate(search, hv, i);
-            pthread_mutex_unlock(&lru_crawler_stats_lock);
-
-            if (hold_lock)
-                item_trylock_unlock(hold_lock);
-            pthread_mutex_unlock(&lru_locks[i]);
-
-            if (crawls_persleep <= 0 && settings.lru_crawler_sleep) {
-                usleep(settings.lru_crawler_sleep);
-                crawls_persleep = settings.crawls_persleep;
-            }
-        }
-    }
-    if (settings.verbose > 2)
-        fprintf(stderr, "LRU crawler thread sleeping\n");
-    STATS_LOCK();
-    stats.lru_crawler_running = false;
-    STATS_UNLOCK();
-    }
-    pthread_mutex_unlock(&lru_crawler_lock);
-    if (settings.verbose > 2)
-        fprintf(stderr, "LRU crawler thread stopping\n");
-
-    return NULL;
-}
-
-static pthread_t item_crawler_tid;
-
-int stop_item_crawler_thread(void) {
-    int ret;
-    pthread_mutex_lock(&lru_crawler_lock);
-    do_run_lru_crawler_thread = 0;
-    pthread_cond_signal(&lru_crawler_cond);
-    pthread_mutex_unlock(&lru_crawler_lock);
-    if ((ret = pthread_join(item_crawler_tid, NULL)) != 0) {
-        fprintf(stderr, "Failed to stop LRU crawler thread: %s\n", strerror(ret));
-        return -1;
-    }
-    settings.lru_crawler = false;
-    return 0;
-}
-
-/* Lock dance to "block" until thread is waiting on its condition:
- * caller locks mtx. caller spawns thread.
- * thread blocks on mutex.
- * caller waits on condition, releases lock.
- * thread gets lock, sends signal.
- * caller can't wait, as thread has lock.
- * thread waits on condition, releases lock
- * caller wakes on condition, gets lock.
- * caller immediately releases lock.
- * thread is now safely waiting on condition before the caller returns.
- */
-int start_item_crawler_thread(void) {
-    int ret;
-
-    if (settings.lru_crawler)
-        return -1;
-    pthread_mutex_lock(&lru_crawler_lock);
-    do_run_lru_crawler_thread = 1;
-    if ((ret = pthread_create(&item_crawler_tid, NULL,
-        item_crawler_thread, NULL)) != 0) {
-        fprintf(stderr, "Can't create LRU crawler thread: %s\n",
-            strerror(ret));
-        pthread_mutex_unlock(&lru_crawler_lock);
-        return -1;
-    }
-    /* Avoid returning until the crawler has actually started */
-    pthread_cond_wait(&lru_crawler_cond, &lru_crawler_lock);
-    pthread_mutex_unlock(&lru_crawler_lock);
-
-    return 0;
-}
-
-/* 'remaining' is passed in so the LRU maintainer thread can scrub the whole
- * LRU every time.
- */
-static int do_lru_crawler_start(uint32_t id, uint32_t remaining) {
-    int i;
-    uint32_t sid;
-    uint32_t tocrawl[3];
-    int starts = 0;
-    tocrawl[0] = id | HOT_LRU;
-    tocrawl[1] = id | WARM_LRU;
-    tocrawl[2] = id | COLD_LRU;
-
-    for (i = 0; i < 3; i++) {
-        sid = tocrawl[i];
-        pthread_mutex_lock(&lru_locks[sid]);
-        if (tails[sid] != NULL) {
-            if (settings.verbose > 2)
-                fprintf(stderr, "Kicking LRU crawler off for LRU %d\n", sid);
-            crawlers[sid].nbytes = 0;
-            crawlers[sid].nkey = 0;
-            crawlers[sid].it_flags = 1; /* For a crawler, this means enabled. */
-            crawlers[sid].next = 0;
-            crawlers[sid].prev = 0;
-            crawlers[sid].time = 0;
-            crawlers[sid].remaining = remaining;
-            crawlers[sid].slabs_clsid = sid;
-            crawler_link_q((item *)&crawlers[sid]);
-            crawler_count++;
-            starts++;
-        }
-        pthread_mutex_unlock(&lru_locks[sid]);
-    }
-    if (starts) {
-        STATS_LOCK();
-        stats.lru_crawler_running = true;
-        stats.lru_crawler_starts++;
-        STATS_UNLOCK();
-        pthread_mutex_lock(&lru_crawler_stats_lock);
-        memset(&crawlerstats[id], 0, sizeof(crawlerstats_t));
-        crawlerstats[id].start_time = current_time;
-        pthread_mutex_unlock(&lru_crawler_stats_lock);
-    }
-    return starts;
-}
-
-static int lru_crawler_start(uint32_t id, uint32_t remaining) {
-    int starts;
-    if (pthread_mutex_trylock(&lru_crawler_lock) != 0) {
-        return 0;
-    }
-    starts = do_lru_crawler_start(id, remaining);
-    if (starts) {
-        pthread_cond_signal(&lru_crawler_cond);
-    }
-    pthread_mutex_unlock(&lru_crawler_lock);
-    return starts;
-}
-
-/* FIXME: Split this into two functions: one to kick a crawler for a sid, and one to
- * parse the string. LRU maintainer code is generating a string to set up a
- * sid.
- * Also only clear the crawlerstats once per sid.
- */
-enum crawler_result_type lru_crawler_crawl(char *slabs) {
-    char *b = NULL;
-    uint32_t sid = 0;
-    int starts = 0;
-    uint8_t tocrawl[MAX_NUMBER_OF_SLAB_CLASSES];
-    if (pthread_mutex_trylock(&lru_crawler_lock) != 0) {
-        return CRAWLER_RUNNING;
-    }
-
-    /* FIXME: I added this while debugging. Don't think it's needed? */
-    memset(tocrawl, 0, sizeof(uint8_t) * MAX_NUMBER_OF_SLAB_CLASSES);
-    if (strcmp(slabs, "all") == 0) {
-        for (sid = 0; sid < MAX_NUMBER_OF_SLAB_CLASSES; sid++) {
-            tocrawl[sid] = 1;
-        }
-    } else {
-        for (char *p = strtok_r(slabs, ",", &b);
-             p != NULL;
-             p = strtok_r(NULL, ",", &b)) {
-
-            if (!safe_strtoul(p, &sid) || sid < POWER_SMALLEST
-                    || sid >= MAX_NUMBER_OF_SLAB_CLASSES-1) {
-                pthread_mutex_unlock(&lru_crawler_lock);
-                return CRAWLER_BADCLASS;
-            }
-            tocrawl[sid] = 1;
-        }
-    }
-
-    for (sid = POWER_SMALLEST; sid < MAX_NUMBER_OF_SLAB_CLASSES; sid++) {
-        if (tocrawl[sid])
-            starts += do_lru_crawler_start(sid, settings.lru_crawler_tocrawl);
-    }
-    if (starts) {
-        pthread_cond_signal(&lru_crawler_cond);
-        pthread_mutex_unlock(&lru_crawler_lock);
-        return CRAWLER_OK;
-    } else {
-        pthread_mutex_unlock(&lru_crawler_lock);
-        return CRAWLER_NOTSTARTED;
-    }
-}
-
-/* If we hold this lock, crawler can't wake up or move */
-void lru_crawler_pause(void) {
-    pthread_mutex_lock(&lru_crawler_lock);
-}
-
-void lru_crawler_resume(void) {
-    pthread_mutex_unlock(&lru_crawler_lock);
-}
-
-int init_lru_crawler(void) {
-    if (lru_crawler_initialized == 0) {
-        memset(&crawlerstats, 0, sizeof(crawlerstats_t) * MAX_NUMBER_OF_SLAB_CLASSES);
-        if (pthread_cond_init(&lru_crawler_cond, NULL) != 0) {
-            fprintf(stderr, "Can't initialize lru crawler condition\n");
-            return -1;
-        }
-        pthread_mutex_init(&lru_crawler_lock, NULL);
-        lru_crawler_initialized = 1;
-    }
-    return 0;
-}
